from typing import Any, Dict, List
from glassgen.sinks.base import BaseSink
from glassgen.sinks.kafka import BaseKafkaClient
import socket

    
class ConfluentKafkaSink(BaseSink, BaseKafkaClient):
    def __init__(self, sink_params: dict):
        self.sink_params = sink_params                
        bootstrap_servers = self.sink_params['bootstrap_servers']
        self.username = self.sink_params['username']
        self.password = self.sink_params['password']
        self.topic = self.sink_params['topic']

        super().__init__(bootstrap_servers)

    def get_client_config(self) -> Dict:
        """Get Confluent client configuration"""
        return {
            "bootstrap.servers": self.bootstrap_servers,
            "security.protocol": self.sink_params['security_protocol'],
            "sasl.mechanisms": self.sink_params['sasl_mechanism'],
            "sasl.username": self.sink_params['sasl_plain_username'],
            "sasl.password": self.sink_params['sasl_plain_password'],
            "client.id": socket.gethostname()
        }
    
    def publish(self, data: Dict[str, Any]) -> None:
        self.send_messages(self.topic, [data])

    def publish_bulk(self, data: List[Dict[str, Any]]) -> None:
        self.send_messages(self.topic, data)

    def close(self) -> None:
        pass


class AivenKafkaSink(BaseSink, BaseKafkaClient):
    def __init__(self, sink_config: dict):
        self.sink_config = sink_config
        bootstrap_servers = self.sink_config['bootstrap_servers']
        self.username = self.sink_config['username']
        self.password = self.sink_config['password']
        self.ca_cert = self.sink_config['ssl_cafile']
        self.topic = self.sink_config['topic']

        super().__init__(bootstrap_servers)
        
    def publish(self, data: Dict[str, Any]) -> None:
        self.send_messages(self.topic, [data])

    def publish_bulk(self, data: List[Dict[str, Any]]) -> None:
        self.send_messages(self.topic, data)

    def close(self) -> None:
        pass

    def get_client_config(self) -> Dict:
        """Get Aiven client configuration""" 
        return {
            "bootstrap.servers": self.bootstrap_servers,
            "security.protocol": self.sink_config['security_protocol'],
            "sasl.mechanisms": self.sink_config['sasl_mechanism'],
            "sasl.username": self.username,
            "sasl.password": self.password,
            "client.id": socket.gethostname(),            
            "ssl.ca.location": self.ca_cert
        }



from typing import Dict, Any, Optional
from pydantic import BaseModel, Field

class SinkConfig(BaseModel):
    type: str
    params: Optional[Dict[str, Any]] = None

class GeneratorConfig(BaseModel):
    rps: int = Field(default=0, ge=0)
    # can be set to -1 to generate indefinitely
    num_records: int = Field(default=100, ge=-1)
    # bulk size for some sinks
    bulk_size: int = Field(default=100, ge=0)

class GlassGenConfig(BaseModel):
    schema_config: Dict[str, Any] = Field(alias="schema")
    sink: SinkConfig
    generator: GeneratorConfig 


import time
from glassgen.schema import BaseSchema
from glassgen.sinks import BaseSink
from glassgen.config import GeneratorConfig

class DynamicBatchController:
    def __init__(self, target_rps: int):
        self.target_rps = target_rps
        self.last_reset = time.time()
        self.records_sent = 0
        self.window = 1.0  # seconds

    def get_batch_size(self, max_batch_size: int = 1000) -> int:
        now = time.time()
        elapsed = now - self.last_reset

        if elapsed >= self.window:
            self._reset(now)

        remaining_time = self.window - elapsed
        remaining_records = max(self.target_rps - self.records_sent, 0)

        if remaining_time <= 0 or remaining_records <= 0:
            self._sleep_until_next_window()
            return self.get_batch_size(max_batch_size)

        est_batch = int(remaining_records * 0.1)
        batch_size = min(est_batch or 1, remaining_records, max_batch_size)
        return max(1, batch_size)

    def record_sent(self, count: int):
        self.records_sent += count
        self._sleep_if_needed()

    def _reset(self, now):
        self.last_reset = now
        self.records_sent = 0

    def _sleep_if_needed(self):
        now = time.time()
        elapsed = now - self.last_reset
        expected_time = self.records_sent / self.target_rps
        if expected_time > elapsed:
            sleep_time = expected_time - elapsed
            time.sleep(sleep_time)

    def _sleep_until_next_window(self):
        now = time.time()
        sleep_time = self.window - (now - self.last_reset)
        if sleep_time > 0:
            time.sleep(sleep_time)
        self._reset(time.time())


class Generator:
    def __init__(self, generator_config: GeneratorConfig, schema: BaseSchema, sink: BaseSink):
        self.generator_config = generator_config
        self.schema = schema
        self.sink = sink
        self.batch_controller = (
            DynamicBatchController(self.generator_config.rps) if self.generator_config.rps > 0 else None
        )
        self.max_bulk_size = 5000

    def _generate_batch(self, num_records: int):
        records = []
        for _ in range(num_records):
            records.append(self.schema._generate_record())
        return records

    def generate(self) -> None:
        """
        Generate records and publish them to the sink.    
        """
        count = 0
        events_to_send = self.generator_config.num_records
        if events_to_send == -1:
            events_to_send = float('inf')
        else:
            events_to_send = int(events_to_send)

        while True:
            batch_size = (
                self.batch_controller.get_batch_size(self.max_bulk_size)
                if self.batch_controller
                else min(self.max_bulk_size, events_to_send - count)
            )
            actual_batch_size = min(batch_size, events_to_send - count)
            records = self._generate_batch(actual_batch_size)
            count += len(records)
            #print(f"Generated {len(records)}. Total records generated: {count} out of {events_to_send}")
            if len(records) > 1:
                self.sink.publish_bulk(records)
            else:
                self.sink.publish(records[0])
            
            if self.batch_controller:
                self.batch_controller.record_sent(actual_batch_size)
                
            if count >= events_to_send:
                break


# a sink that prints the data to the console
from glassgen.sinks import BaseSink

class PrintSink(BaseSink):
    def __init__(self, config=None):
        self.config = config
    
    def publish(self, data):
        print(data)
    
    def close(self):
        pass


from typing import Dict, Any, Union
import time
from glassgen.schema import BaseSchema
from glassgen.schema.schema import ConfigSchema
from glassgen.sinks import SinkFactory, BaseSink
from glassgen.generator import Generator
from glassgen.config import GlassGenConfig


def generate(config: Union[Dict[str, Any], GlassGenConfig], schema: BaseSchema=None, sink: BaseSink = None) -> None:
    """
    Generate data based on the provided configuration.
    
    Args:
        config: Configuration dictionary or GlassGenConfig object
        schema: Optional schema object to use for generating data
        sink: Optional sink object to use for sending generated data
    """
    # Convert dict to Pydantic model if needed
    if isinstance(config, dict):
        config = GlassGenConfig(**config)
    
    # Create schema if not provided
    if schema is None:
        schema = ConfigSchema.from_dict(config.schema_config)
        schema.validate()
    
    # Create sink if not provided
    if sink is None:
        sink = SinkFactory.create(config.sink.type, config.sink.params) 
    
    # Create and run generator
    start_time = time.time()
    generator = Generator(config.generator, schema, sink)
    generator.generate()
    end_time = time.time()
    time_taken_ms = round((end_time - start_time)*1000)
    #print(f"Time taken: {time_taken_ms} milliseconds")
    print(f"Generated {config.generator.num_records} records and published to {config.sink.type} in {time_taken_ms} milliseconds")

    # Close sink
    sink.close()


# usage of a custom sink
import glassgen
import json
from glassgen.schema import UserSchema

with open("config.kafka.json") as f:
    config_json = json.load(f)

glassgen.generate(config=config_json, schema=UserSchema())


from glassgen.sinks.base import BaseSink
from glassgen.sinks.csv_sink import CSVSink
from glassgen.sinks.kafka_sink import ConfluentKafkaSink
from glassgen.sinks.kafka_sink import AivenKafkaSink
from glassgen.sinks.webhook_sink import WebHookSink

__all__ = ["BaseSink", "CSVSink", "SinkFactory", "ConfluentKafkaSink", "AivenKafkaSink", "WebHookSink"]


class SinkFactory:
    _sinks = {
        "csv": CSVSink,
        "kafka.confluent": ConfluentKafkaSink,
        "kafka.aiven": AivenKafkaSink,
        "webhook": WebHookSink,
    }

    @classmethod
    def create(cls, class_type: str, sink_config: dict):
        if class_type not in cls._sinks:
            raise ValueError(f"Unknown class_type: {class_type}")
        return cls._sinks[class_type](sink_config)



# usage of a custom sink
from print_sink import PrintSink
import glassgen
import json

sink = PrintSink()

with open("config.json") as f:
    config_json = json.load(f)

glassgen.generate(config=config_json, sink=sink)